package cp213;

import java.awt.Component;
import java.awt.Dimension;
import java.awt.Point;
import java.awt.Toolkit;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.Future;

import javax.swing.BoxLayout;
import javax.swing.JButton;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.WindowConstants;

// -------------------------------------------------------------------------------
/**
 * Provides a way of spawning CountModel objects and their views.
 *
 * @author David Brown
 * @version 2017-11-15
 *
 *          Spawns counter frames when the "Spawn" button is pushed.
 */
@SuppressWarnings("serial")
public class SpawnView extends JPanel {

    // -------------------------------------------------------------------------------
    /**
     * Kills the model's thread when the view window is closed/disposed.
     */
    private class DisposeListener extends WindowAdapter {

	// Refers to the currently running thread.
	private Future<?> future = null;

	/**
	 * Add the Future object to the listener.
	 *
	 * @param model
	 */
	public DisposeListener(final Future<?> future) {
	    this.future = future;
	}

	@Override
	public void windowClosing(final WindowEvent e) {
	    // Cancel the currently running thread if it is not already
	    // cancelled.
	    if (this.future != null && !this.future.isCancelled()) {
		this.future.cancel(true);
	    }
	}
    }

    // -------------------------------------------------------------------------------
    /**
     * Spawns a new CountModel frame. Each counter frame is given a unique
     * number based upon the static {@code count} attribute.
     */
    private class SpawnListener implements ActionListener {

	@Override
	public void actionPerformed(final ActionEvent evt) {
	    final CountModel model = new CountModel();
	    final JFrame frame = new JFrame("Counter " + SpawnView.count++);
	    frame.setContentPane(new CountView(model));
	    frame.setSize(SIZE);
	    frame.setDefaultCloseOperation(WindowConstants.DISPOSE_ON_CLOSE);
	    // Update the location of the next counter frame.
	    SpawnView.this.nextLocation();
	    frame.setLocation(SpawnView.this.location);

	    // Execute the CountModel with a thread from the thread pool.

	    // The Future object monitors the thread.
	    final Future<?> future = SpawnView.this.threadPool.submit(model);

	    frame.addWindowListener(new DisposeListener(future));
	    frame.setVisible(true);
	}
    }

    // -------------------------------------------------------------------------------
    // Define the size and location of the counter frames.
    public static final int WIDTH = 200;
    public static final int HEIGHT = 100;
    public static final Dimension SIZE = new Dimension(WIDTH, HEIGHT);

    private final Point location = new Point(0, 0);
    // Attributes to handle screen placement.
    private final Dimension screenSize = Toolkit.getDefaultToolkit()
	    .getScreenSize();
    // Tracks the number of counter frames spawned.
    private static int count = 0;
    // Spawns new counter frames when pressed.
    private final JButton spawnButton = new JButton("Spawn");

    // Provides an unbounded pool of execution threads.
    // private final ExecutorService threadPool =
    // Executors.newCachedThreadPool();
    // Or: Provides a limited number of threads.
    private final ExecutorService threadPool = Executors.newFixedThreadPool(4);

    // -------------------------------------------------------------------------------
    /**
     * Constructor.
     */
    public SpawnView() {
	this.layoutView();
	this.registerListeners();
    }

    // -------------------------------------------------------------------------------
    /**
     * Lays out the panel components.
     */
    private void layoutView() {
	this.setLayout(new BoxLayout(this, BoxLayout.Y_AXIS));
	final JLabel instructLabel = new JLabel("Spawn a Counter");
	instructLabel.setAlignmentX(Component.CENTER_ALIGNMENT);
	instructLabel.setAlignmentY(Component.CENTER_ALIGNMENT);
	this.add(instructLabel);
	this.spawnButton.setAlignmentX(Component.CENTER_ALIGNMENT);
	this.spawnButton.setAlignmentY(Component.CENTER_ALIGNMENT);
	this.add(this.spawnButton);
    }

    // -------------------------------------------------------------------------------
    /**
     * Determines the location of the next counter frame. Lays out frames from
     * left-to-right, top-to-bottom of the current screen.
     */
    private void nextLocation() {
	int x = this.location.x + WIDTH;
	int y = this.location.y;

	if (x + WIDTH > this.screenSize.width) {
	    x = 0;
	    y += HEIGHT;
	}
	this.location.setLocation(x, y);
    }

    // -------------------------------------------------------------------------------
    /**
     * Registers all listeners with the appropriate inner listener classes.
     */
    private void registerListeners() {
	this.spawnButton.addActionListener(new SpawnListener());
    }
}
